const path = require('path');
const stylelint = require('stylelint');
const { report, ruleMessages, validateOptions } = stylelint.utils;
const valueParser = require('postcss-value-parser');
const sdsNamespaces = require('@salesforce-ux/sds-namespaces');

// Plugin name
const ruleName = 'sds-stylelint-plugin/styling-hooks-pattern';

// Custom message for report
const messages = ruleMessages(ruleName, {
  expected: (prop, report, message) => {
    const errors = report
      .map((error, key) => {
        const expected = error.info.expected
          ? `Expected ${error.group} value(s) of "${error.info.expected}" but received "${error.info.received}".`
          : '';
        return `${key + 1}.) ${expected}`;
      })
      .join(' ');
    const error = `${report.length} found on ${prop} fails naming pattern. ${errors}${
      message && message.length > 0 ? `[Analysis Warning]: ${message.map((error) => error).join(' ')}` : ''
    }`;
    return error;
  },
});

// Metadata
const meta = {
  global: {
    root: [],
    shared: ['button', 'container', 'content', 'heading', 'input', 'link', 'popup'],
  },
  valid: {
    scope: ['g', 's', 'c'],
    category: ['color', 'font', 'radius', 'shadow', 'sizing', 'spacing'],
    property: [
      'background',
      'foreground',
      'border',
      'link',
      'style',
      'size',
      'weight',
      'lineheight',
      'decoration',
      'align',
      'inline',
      'block',
      'square',
      'width',
      'height',
    ],
    attribute: ['start', 'startstart', 'startend', 'end', 'endstart', 'endend', 'gap'],
    state: [
      'checked',
      'pressed',
      'disabled',
      'readonly',
      'valid',
      'invalid',
      'indeterminate',
      'error',
      'placeholder',
      'selected',
    ],
    psuedoState: ['hover', 'focus', 'active', 'visited'],
  },
  restricted: [''],
};

/**
 * Store keys based on metadata
 * @param {Object[]} arr An array of words from the css custom property
 * @returns {Object} An object of keys based on metadata
 */
const findIndexes = (arr) => {
  const local = {
    scope: null,
    context: null,
    variant: null,
    element: null,
    category: null,
    property: null,
    attribute: null,
    state: null,
    psuedoState: null,
  };

  arr.map((el, key) => {
    meta.valid.scope.find((val) => {
      if (el === val) {
        local.scope = key;
      }
    });
    meta.valid.category.find((val) => {
      if (el === val) {
        local.category = key;
      }
    });
    meta.valid.property.find((val) => {
      if (el === val) {
        local.property = key;
      }
    });
    meta.valid.attribute.find((val) => {
      if (el === val) {
        local.attribute = key;
      }
    });
    meta.valid.state.find((val) => {
      if (el === val) {
        local.state = key;
      }
    });
    meta.valid.psuedoState.find((val) => {
      if (el === val) {
        local.psuedoState = key;
      }
    });
  });
  return local;
};

/**
 * Validate namespace
 * @param {String} ns Namespace to be validated against the sds-namespace definitions
 * @returns {Object} Validation object
 */
const validateNs = (ns) => {
  for (const team of sdsNamespaces) {
    if (ns === team.name) {
      return { valid: true, received: ns };
    }
  }
  return { valid: false, expected: sdsNamespaces.map((team) => team.name), received: ns };
};

/**
 * Validate scope
 * @param {Object[]} arr An array of words from the css custom property
 * @param {Object} keys An object of keys based on metadata
 * @returns {Object} Validation object
 */
const validateScope = (arr, keys) => {
  const scope = arr[keys.scope];
  let isGlobal = false;
  let isShared = false;
  let isComponent = false;
  if (scope && scope === 'c') isComponent = true;
  if (scope && scope === 'g') isGlobal = true;
  if (scope && scope === 's') isShared = true;
  if (arr.includes(scope)) return { valid: true, scope, isGlobal, isShared, isComponent };
  return { valid: false, expected: meta.valid.scope, received: arr[1] };
};

/**
 * Validate context
 * @param {Object[]} arr An array of words from the css custom property
 * @param {String} location The file system path
 * @param {Object} keys An object of keys based on metadata
 * @returns {Object} Validation object
 */
const validateContext = (arr, location, keys) => {
  const scope = validateScope(arr, keys);
  // Component scope checks & Shared scope checks
  if (
    (location && location.split('-').join('') === arr[2]) ||
    (location && location.split('-')[0] === arr[2]) ||
    (scope.isShared && meta.global.shared.includes(arr[2]))
  ) {
    // Since we know where the context is coming from, lets update the keys to help validate other elements
    keys.context = 2;
    return { valid: true, received: arr[2] };
    // Skip if global since we dont require context at the moment
  } else if (scope.isGlobal) {
    return { skip: true };
  } else {
    keys.context = 2;
    return {
      valid: false,
      expected: scope.isComponent ? location.split('-').join('') : scope.isShared ? meta.global.shared : null,
      received: arr[2],
    };
  }
};

/**
 * Validate the existence of an element
 * @param {Object} obj The object from the first Promise that has valuable information about the custom prop
 * @returns {Object} Validation object or original object to pass along to next function or definition to skip if no element
 */
const validateElement = (obj) => {
  // If the context has a key of 2 and the category has a key of 4, we know its an element
  if (obj.fuzzyKeys.has(obj.keys.context + 1) && obj.keys.category === 4) {
    obj.keys.element = 3;
    obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.element);
    return { valid: true, received: obj.groups[obj.keys.element] };
  } else if (obj.fuzzyKeys.has(obj.keys.context + 1) && obj.keys.category === 5) {
    obj.keys.element = 4;
    obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.element);
    return { valid: true, received: obj.groups[obj.keys.element] };
  }
  // Skip because we can't determine an element exists
  else {
    return { skip: true };
  }
};

const validateVariant = (obj) => {
  if (obj.fuzzyKeys.has(obj.keys.context + 1) && obj.keys.element === 4 && obj.keys.category === 5) {
    obj.keys.variant = 3;
    obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.variant);
    return { valid: true, received: obj.groups[obj.keys.variant] };
  }
  // Skip because we can't determine a variant exists
  return { skip: true };
};

/**
 * Validate Category
 * @param {Object} obj The object from the first Promise that has valuable information about the custom prop
 * @todo Skip scope validation based on stylelint option
 * @returns {Object} Validation object
 */
const validateCategory = (obj) => {
  if (obj.validate.scope.isGlobal) return { skip: true };
  const results = analyzeFuzzy(obj).filter((item) => item.word);

  if (obj.keys.category) {
    return { valid: true, received: obj.groups[obj.keys.category] };
  } else if (obj.fuzzyKeys.size > 0) {
    let valid;
    results.map((item) => {
      if (!item.lastKey && obj.keys.context === 2 && obj.keys.property === 4) {
        obj.keys.category = obj.keys.context + 1;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.category);
        valid = {
          valid: false,
          expected: meta.valid.category,
          received: obj.groups[obj.keys.category],
        };
      } else if (item.lastKey && results.length === 1) {
        obj.keys.category = obj.keys.context + 1;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.category);
        valid = {
          valid: false,
          expected: meta.valid.category,
          received: obj.groups[obj.keys.category],
        };
      }
      // check for element or variant
      else if (!obj.keys.category && item.lastKey && results.length === 2 && item.key !== 3) {
        obj.keys.category = obj.keys.context + 2;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.category);
        valid = { valid: false, received: obj.groups[item.key] };
        obj.customMessage.push(
          `We received "${
            obj.groups[item.key]
          }" and have low confidence in determining the category. Possible values are ${
            meta.valid.category
          }.`,
        );
      } else if (!obj.keys.category && !item.lastKey && results.length === 2 && item.key !== 3) {
        obj.keys.category = obj.keys.context + 2;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.category);
        valid = { valid: false, received: obj.groups[item.key] };
        obj.customMessage.push(
          `We received "${
            obj.groups[item.key]
          }" and have low confidence in determining the category. Possible values are ${
            meta.valid.category
          }.`,
        );
      } else if (!obj.keys.category && results.length >= 3 && item.key !== 3 && item.key !== 4) {
        obj.keys.category = obj.keys.context + 3;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.category);
        valid = {
          valid: false,
          expected: meta.valid.category,
          received: obj.groups[obj.keys.category],
        };
      }
    });
    return valid;
  }
};

/**
 * Validate Property
 * @param {Object} obj The object from the first Promise that has valuable information about the custom prop
 * @todo Skip scope validation based on stylelint option
 * @returns {Object} Validation object or definition to skip if not found
 */
const validateProperty = (obj) => {
  if (obj.validate.scope.isGlobal) return { skip: true };

  const results = analyzeFuzzy(obj).filter((item) => item.word);

  // If key has already been stored, we know its valid
  if (obj.keys.property) {
    return { valid: true, received: obj.groups[obj.keys.property] };
  }
  // Check that something exists after the category key while ensuring its not an attribute, state, or psuedoState
  else if (obj.fuzzyKeys.size > 0 && results.length > 0) {
    let valid;

    results.map((item) => {
      if (obj.keys.category && item.key === obj.keys.category + 1) {
        obj.keys.property = obj.keys.category + 1;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.property);
        valid = { valid: false, expected: meta.valid.property, received: obj.groups[obj.keys.property] };
      }
    });
    return valid;
  } else {
    return { skip: true };
  }
};

/**
 * Validate Attribute
 * @param {Object} obj The object from the first Promise that has valuable information about the custom prop
 * @todo Skip scope validation based on stylelint option
 * @returns {Object} Validation object or definition to skip if not found
 */
const validateAttribute = (obj) => {
  if (obj.validate.scope.isGlobal) return { skip: true };

  const results = analyzeFuzzy(obj).filter((item) => item.word);

  // If key has already been stored, we know its valid
  if (obj.keys.attribute) {
    return { valid: true, received: obj.groups[obj.keys.attribute] };
  } else if (obj.fuzzyKeys.size > 0 && results.length > 0) {
    let valid;
    results.map((item) => {
      if (
        (obj.keys.property && obj.keys.state === item.key + 1) ||
        (obj.keys.property && obj.keys.psuedoState === item.key + 1)
      ) {
        obj.keys.attribute = obj.keys.property + 1;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.attribute);
        valid = {
          valid: false,
          expected: meta.valid.attribute,
          received: obj.groups[obj.keys.attribute],
        };
      } else if (
        !obj.keys.attribute &&
        results.length === 3 &&
        obj.keys.property &&
        item.key === obj.keys.property + 1
      ) {
        obj.keys.attribute = obj.keys.property + 1;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.attribute);
        valid = {
          valid: false,
          expected: meta.valid.attribute,
          received: obj.groups[obj.keys.attribute],
        };
      } else if (!obj.keys.attribute && item.lastKey && item.key === obj.keys.property + 1) {
        valid = { valid: false, received: obj.groups[item.key] };
        obj.customMessage.push(
          `We received "${
            obj.groups[item.key]
          }" and have low confidence in determining if this is an attribute, state, or psuedoState. Possible values are ${
            meta.valid.attribute
          },${meta.valid.state},${meta.valid.psuedoState}.`,
        );
      } else if (
        !obj.keys.attribute &&
        !item.lastKey &&
        item.key === obj.keys.property + 1 &&
        results.length <= 2
      ) {
        valid = { valid: false, received: obj.groups[item.key] };
        obj.customMessage.push(
          `We received "${
            obj.groups[item.key]
          }" and have medium confidence that this is an attribute but might also be a state, or psuedoState. Possible values are ${
            meta.valid.attribute
          },${meta.valid.state},${meta.valid.psuedoState}.`,
        );
      } else if (obj.keys.property !== obj.keys.category + 1) {
        valid = { valid: false, received: obj.groups[item.key] };
        obj.customMessage.push(
          `We received "${obj.groups[item.key]}" but we do not allow words in between "${
            obj.groups[obj.keys.category]
          }", a category, and "${obj.groups[obj.keys.property]}", a property.`,
        );
      }
    });
    return valid;
  } else {
    return { skip: true };
  }
};

/**
 * Validate state
 * @param {Object} obj The object from the first Promise that has valuable information about the custom prop
 * @todo Skip scope validation based on stylelint option
 * @returns {Object} Validation object or definition to skip if not found
 */
const validateState = (obj) => {
  if (obj.validate.scope.isGlobal) return { skip: true };

  const results = analyzeFuzzy(obj).filter((item) => item.word);

  if (obj.keys.state) {
    return { valid: true, received: obj.groups[obj.keys.state] };
  } else if (obj.fuzzyKeys.size > 0 && results.length > 0) {
    let valid;
    results.map((item) => {
      if (!item.lastKey && item.key === obj.keys.attribute + 1) {
        obj.keys.state = obj.keys.attribute + 1;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.state);
        valid = {
          valid: false,
          expected: meta.valid.state,
          received: obj.groups[obj.keys.state],
        };
      } else if (!obj.keys.state && item.lastKey && results.length === 2) {
        valid = { valid: false, received: obj.groups[item.key] };
        obj.customMessage.push(
          `We received "${
            obj.groups[item.key]
          }" and have medium confidence that this could be a state or a psuedoState. Possible values are ${
            meta.valid.state
          },${meta.valid.psuedoState}.`,
        );
      }
      // fuzzy result already captured previously
      else if (
        (item.lastKey && obj.customMessage.length === 1) ||
        (!obj.validate.attribute.valid && obj.customMessage.length === 1)
      ) {
        valid = { skip: true };
      }
    });
    return valid;
  } else {
    return { skip: true };
  }
};

/**
 * Validate state
 * @param {Object} obj The object from the first Promise that has valuable information about the custom prop
 * @todo Skip scope validation based on stylelint option
 * @returns {Object} Validation object or definition to skip if not found
 */
const validatePsuedoState = (obj) => {
  if (obj.validate.scope.isGlobal) return { skip: true };
  const results = analyzeFuzzy(obj).filter((item) => item.word);

  if (obj.keys.psuedoState) {
    return { valid: true, received: obj.groups[obj.keys.psuedoState] };
  } else if (obj.fuzzyKeys.size > 0 && results.length > 0) {
    let valid;
    results.map((item) => {
      // fuzzy result already captured previously
      if (item.lastKey && obj.customMessage.length === 1) {
        valid = { skip: true };
      } else if (item.lastKey && results.length === 2) {
        valid = { skip: true };
      } else if (
        (item.lastKey && item.key === obj.keys.attribute + 1) ||
        (item.lastKey && item.key === obj.keys.state + 1)
      ) {
        obj.keys.psuedoState = item.key;
        obj.fuzzyKeys = removeFromSet(obj.fuzzyKeys, obj.keys.psuedoState);
        valid = {
          valid: false,
          expected: meta.valid.psuedoState,
          received: obj.groups[obj.keys.psuedoState],
        };
      } else {
        valid = { skip: true };
      }
    });
    return valid;
  } else {
    return { skip: true };
  }
};

const analyzeFuzzy = (obj) => {
  const suggested = [];
  const checkFuzzyKeys = obj.fuzzyKeys.values();

  obj.fuzzyKeys.forEach((it) => {
    const suggest = {
      word: null,
      stored: null,
      key: null,
      lastKey: null,
    };
    if (obj.groups[checkFuzzyKeys.next().value]) {
      if (!Object.values(obj.keys).includes(it)) {
        suggest.word = obj.groups[it];
        suggest.stored = Object.values(obj.keys).includes(it);
        suggest.key = it;
        suggest.lastKey = it === obj.lastKey;
      }
    }
    suggested.push(suggest);
  });
  return suggested;
};

/**
 * This is for Jest, it gets the context from the class selector
 * @param {*} str
 * @returns string
 */
const getContextFromString = (str) => {
  const regEx = RegExp(/([^\W]+)/, 'g');
  return regEx.exec(str)[0];
};

/**
 * Retrieve missing keys from initial index storage based on metadata
 * @param {*} set
 * @param {*} total
 * @returns {Object[]} Set of unused keys
 */
const getMissingKeys = (set, total) => {
  let keySet = [];
  Object.keys(set).map((key) => {
    if (set[key]) {
      keySet.push(set[key]);
    }
  });
  if (!keySet.includes(total)) {
    keySet = [...keySet, total];
  }
  const range = Array.from(Array(Math.max(...keySet)).keys());
  const missing = range
    .map((i) => {
      if (keySet.indexOf(i) < 0 && i > Math.min(...keySet)) {
        return i;
      }
    })
    .filter((f) => f);
  return missing;
};

const removeFromArray = (arr, value) => {
  return arr.filter((el) => el !== value);
};

const removeFromSet = (set, value) => {
  set.delete(value);
  return set;
};

module.exports = stylelint.createPlugin(ruleName, function (primary, options) {
  return function (root, result) {
    const validOptions = validateOptions(result, ruleName, { actual: primary });
    if (!validOptions) return;
    if (options && options.valid) {
      Object.keys(options.valid).map((group) => {
        if (Object.getOwnPropertyNames(meta.valid).includes(group)) {
          const merged = new Set([...meta.valid[group], ...options.valid[group]]);
          const newOptions = [...Array.from(merged)];
          meta.valid[group] = newOptions;
        }
      });
    }

    root.walkDecls((decl) => {
      const parsedValue = valueParser(decl.value);

      parsedValue.walk((node) => {
        const context = decl.source.input.file
          ? path.parse(decl.source.input.file).name.split('.')[0]
          : getContextFromString(decl.parent.selector);
        if (node.type !== 'word') return;
        if (node.value.startsWith('--')) {
          // Break words of custom prop into groups
          const groups = node.value.match(/\b[a-z0-9]+\b/g);
          // Store keys based on metadata
          const keys = findIndexes(groups);
          // Store remaining unused keys, this is used to determine validity of optional words
          const fuzzyKeys = new Set(removeFromArray(getMissingKeys(keys, groups.length), 2));
          // Populate pattern object to check against custom prop
          const pattern = new Promise((resolve, reject) => {
            resolve({
              prop: node.value,
              groups: groups,
              keys,
              lastKey: groups.length - 1,
              fuzzyKeys,
              // Known validations
              validate: {
                ns: validateNs(groups[0]),
                scope: validateScope(groups, keys),
                context: validateContext(groups, context, keys),
              },
              customMessage: [],
            });
          });

          // Required validations
          pattern.then((result) => {
            result.validate.category = validateCategory(result);
          });

          // Optional validations
          pattern.then((result) => {
            result.validate.element = validateElement(result);
            result.validate.variant = validateVariant(result);
            result.validate.property = validateProperty(result);
            result.validate.attribute = validateAttribute(result);
            result.validate.state = validateState(result);
            result.validate.psuedoState = validatePsuedoState(result);
            // console.log(result);
            return result;
          });

          // Check validity on Promise and send back invalid items
          const onInvalid = pattern.then((result) => {
            let isValid = true;
            const invalid = Object.keys(result.validate).map((valid) => {
              // if (result.validate[valid] === undefined) console.warn(result);
              if (result.validate[valid] && result.validate[valid].valid === false) {
                isValid = false;
                return {
                  group: valid,
                  info:
                    result.customMessage.length > 0
                      ? Object.assign(result.validate[valid], { message: result.customMessage })
                      : result.validate[valid],
                };
              }
            });
            const invalidSet = new Set(invalid);
            return isValid
              ? false
              : {
                  prop: result.prop,
                  result: [...invalidSet].filter((x) => x !== undefined),
                  message: result.customMessage,
                };
          });

          // Throw report if invalid items show up in the onInvalid Set
          onInvalid.then((invalid) => {
            if (invalid) {
              report({
                ruleName,
                result: result,
                message: messages.expected(invalid.prop, invalid.result, invalid.message),
                node: decl,
              });
            }
          });
        }
      });
    });
  };
});

module.exports.ruleName = ruleName;
module.exports.messages = messages;
module.exports.meta = meta;
